@@ -36,7 +36,7 @@ def signal_handling(signum, frame):
 
 signal.signal(signal.SIGINT, signal_handling)
 
-perceptor, normalize_image = load()
+perceptor, normalize_image = load('ViT-B/32', jit = False)
 
 # Helpers
 
@@ -47,6 +47,7 @@ def exists(val):
 def default(val, d):
     return val if exists(val) else d
 
+
 def interpolate(image, size):
     return F.interpolate(image, (size, size), mode='bilinear', align_corners=False)
 
@@ -85,7 +86,7 @@ def open_folder(path):
 
 
 def norm_siren_output(img):
-    return ((img + 1) * 0.5).clamp(0, 1)
+    return ((img + 1) * 0.5).clamp(0.0, 1.0)
 
 
 def create_clip_img_transform(image_width):
@@ -101,6 +102,19 @@ def create_clip_img_transform(image_width):
     return transform
 
 
+def create_text_path(text=None, img=None, encoding=None):
+    if text is not None:
+        input_name = text.replace(" ", "_")[:perceptor.context_length]
+    elif img is not None:
+        if isinstance(img, str):
+            input_name = "".join(img.replace(" ", "_").split(".")[:-1])
+        else:
+            input_name = "PIL_img"
+    else:
+        input_name = "your_encoding"
+    return input_name
+
+
 class DeepDaze(nn.Module):
     def __init__(
             self,
@@ -110,7 +124,10 @@ class DeepDaze(nn.Module):
             image_width=512,
             loss_coef=100,
             theta_initial=None,
-            theta_hidden=None
+            theta_hidden=None,
+            lower_bound_cutout=0.1, # should be smaller than 0.8
+            upper_bound_cutout=1.0,
+            saturate_bound=False,
     ):
         super().__init__()
         # load clip
@@ -141,7 +158,11 @@ class DeepDaze(nn.Module):
             image_height=image_width
         )
 
-        self.generate_size_schedule()
+        self.saturate_bound = saturate_bound
+        self.saturate_limit = 0.75 # cutouts above this value lead to destabilization
+        self.lower_bound_cutout = lower_bound_cutout
+        self.upper_bound_cutout = upper_bound_cutout
+
 
     def forward(self, text_embed, return_loss=True, dry_run=False):
         out = self.model()
@@ -150,84 +171,29 @@ class DeepDaze(nn.Module):
         if not return_loss:
             return out
 
-        pieces = []
+        # sample cutout sizes between lower and upper bound
         width = out.shape[-1]
-        size_slice = slice(self.num_batches_processed, self.num_batches_processed + self.batch_size)
-
-        for size in self.scheduled_sizes[size_slice]:
-            apper = rand_cutout(out, size)
-            apper = interpolate(apper, 224)
-            pieces.append(normalize_image(apper))
-
-        image = torch.cat(pieces)
+        lower_bound = self.lower_bound_cutout
+        if self.saturate_bound:
+            progress_fraction = self.num_batches_processed / self.total_batches
+            lower_bound += (self.saturate_limit - self.lower_bound_cutout) * progress_fraction
+            
+        lower = lower_bound * width
+        upper = self.upper_bound_cutout * width
+        sizes = torch.randint(int(lower), int(upper), (self.batch_size,))
 
+        # create normalized random cutouts
+        image_pieces = torch.cat([normalize_image(interpolate(rand_cutout(out, size), 224)) for size in sizes])
+        # calc image embedding
         with autocast(enabled=False):
-            image_embed = perceptor.encode_image(image)
-
+            image_embed = perceptor.encode_image(image_pieces)
+        # count batches
         if not dry_run:
             self.num_batches_processed += self.batch_size
-
+        # calc loss
         loss = -self.loss_coef * torch.cosine_similarity(text_embed, image_embed, dim=-1).mean()
-        return loss
+        return out, loss
 
-    def generate_size_schedule(self):
-        batches = 0
-        counter = 0
-        self.scheduled_sizes = []
-
-        while batches <= self.total_batches:
-            counter += 1
-            sizes = self.sample_sizes(counter)
-            batches += len(sizes)
-            self.scheduled_sizes.extend(sizes)
-
-    def sample_sizes(self, counter):
-        pieces_per_group = 4
-
-        # 6 piece schedule increasing in context as model saturates
-        if counter < 500:
-            partition = [4, 5, 3, 2, 1, 1]
-        elif counter < 1000:
-            partition = [2, 5, 4, 2, 2, 1]
-        elif counter < 1500:
-            partition = [1, 4, 5, 3, 2, 1]
-        elif counter < 2000:
-            partition = [1, 3, 4, 4, 2, 2]
-        elif counter < 2500:
-            partition = [1, 2, 2, 4, 4, 3]
-        elif counter < 3000:
-            partition = [1, 1, 2, 3, 4, 5]
-        else:
-            partition = [1, 1, 1, 2, 4, 7]
-
-        dbase = .38
-        step = .1
-        width = self.image_width
-
-        sizes = []
-        for part_index in range(len(partition)):
-            groups = partition[part_index]
-            for _ in range(groups * pieces_per_group):
-                sizes.append(torch.randint(
-                    int((dbase + step * part_index + .01) * width),
-                    int((dbase + step * (1 + part_index)) * width), ()))
-
-        sizes.sort()
-        return sizes
-
-
-    
-def create_text_path(text=None, img=None, encoding=None):
-    if text is not None:
-        input_name = text.replace(" ", "_")
-    elif img is not None:
-        if isinstance(img, str):
-            input_name = "".join(img.replace(" ", "_").split(".")[:-1])
-        else:
-            input_name = "PIL_img"
-    else:
-        input_name = "your_encoding"
-    return input_name
 
 class Imagine(nn.Module):
     def __init__(
@@ -244,7 +210,7 @@ class Imagine(nn.Module):
             num_layers=16,
             epochs=20,
             iterations=1050,
-            save_progress=False,
+            save_progress=True,
             seed=None,
             open_folder=True,
             save_date_time=False,
@@ -253,6 +219,12 @@ class Imagine(nn.Module):
             start_image_lr=3e-4,
             theta_initial=None,
             theta_hidden=None,
+            lower_bound_cutout=0.1, # should be smaller than 0.8
+            upper_bound_cutout=1.0,
+            saturate_bound=False,
+            create_story=False,
+            story_start_words=5,
+            story_words_per_epoch=5,
     ):
 
         super().__init__()
@@ -263,19 +235,37 @@ class Imagine(nn.Module):
             torch.cuda.manual_seed(seed)
             random.seed(seed)
             torch.backends.cudnn.deterministic = True
-
-        self.epochs = epochs
+            
+        # fields for story creation:
+        self.create_story = create_story
+        self.words = None
+        self.all_words = text.split(" ") if text is not None else None
+        self.num_start_words = story_start_words
+        self.words_per_epoch = story_words_per_epoch
+        if create_story:
+            assert text is not None,  "We need text input to create a story..."
+            # overwrite epochs to match story length
+            num_words = len(self.all_words)
+            self.epochs = 1 + (num_words - self.num_start_words) / self.words_per_epoch
+            # add one epoch if not divisible
+            self.epochs = int(self.epochs) if int(self.epochs) == self.epochs else int(self.epochs) + 1
+            print("Running for ", self.epochs, "epochs")
+        else: 
+            self.epochs = epochs
+        
         self.iterations = iterations
         self.image_width = image_width
-        total_batches = epochs * iterations * batch_size * gradient_accumulate_every
-
+        total_batches = self.epochs * self.iterations * batch_size * gradient_accumulate_every
         model = DeepDaze(
             total_batches=total_batches,
             batch_size=batch_size,
             image_width=image_width,
             num_layers=num_layers,
             theta_initial=theta_initial,
-            theta_hidden=theta_hidden
+            theta_hidden=theta_hidden,
+            lower_bound_cutout=lower_bound_cutout,
+            upper_bound_cutout=upper_bound_cutout,
+            saturate_bound=saturate_bound,
         ).cuda()
 
         self.model = model
@@ -290,9 +280,9 @@ class Imagine(nn.Module):
         self.image = img
         self.textpath = create_text_path(text=text, img=img, encoding=clip_encoding)
         self.filename = self.image_output_path()
-
+        
         # create coding to optimize for
-        self.clip_img_transform = create_clip_img_transform(perceptor.input_resolution.item())
+        self.clip_img_transform = create_clip_img_transform(224)
         self.clip_encoding = self.create_clip_encoding(text=text, img=img, encoding=clip_encoding)
 
         self.start_image = None
@@ -303,41 +293,68 @@ class Imagine(nn.Module):
             assert file.exists(), f'file does not exist at given starting image path {self.start_image_path}'
             image = Image.open(str(file))
 
-            transform = T.Compose([
-                T.Resize(image_width),
-                T.CenterCrop((image_width, image_width)),
-                T.ToTensor(),
-                T.Normalize(0.5, 0.5)
-            ])
-
-            image_tensor = transform(image)[None, ...].cuda()
+            image_tensor = self.clip_img_transform(image)[None, ...].cuda()
             self.start_image = image_tensor
             
     def create_clip_encoding(self, text=None, img=None, encoding=None):
         self.text = text
         self.img = img
         if encoding is not None:
-            return encoding.cuda()
+            encoding = encoding.cuda()
+        elif self.create_story:
+            encoding = self.update_story_encoding(epoch=0, iteration=1)
+        elif text is not None and img is not None:
+            encoding = (self.create_text_encoding(text) + self.create_img_encoding(img)) / 2
         elif text is not None:
-            return self.create_text_encoding(text)
+            encoding = self.create_text_encoding(text)
         elif img is not None:
-            return self.create_img_encoding(img)
+            encoding = self.create_img_encoding(img)
+        return encoding
 
     def create_text_encoding(self, text):
         tokenized_text = tokenize(text).cuda()
-        text_encoding = perceptor.encode_text(tokenized_text).detach()
+        with torch.no_grad():
+            text_encoding = perceptor.encode_text(tokenized_text).detach()
         return text_encoding
     
     def create_img_encoding(self, img):
         if isinstance(img, str):
             img = Image.open(img)
         normed_img = self.clip_img_transform(img).unsqueeze(0).cuda()
-        img_encoding = perceptor.encode_image(normed_img).detach()
+        with torch.no_grad():
+            img_encoding = perceptor.encode_image(normed_img).detach()
         return img_encoding
     
     def set_clip_encoding(self, text=None, img=None, encoding=None):
-        encoding = self.create_clip_encoding(text=text, img=img, encoding=encoding).cuda()
-        self.clip_encoding = encoding
+        encoding = self.create_clip_encoding(text=text, img=img, encoding=encoding)
+        self.clip_encoding = encoding.cuda()
+        
+    def update_story_encoding(self, epoch, iteration):
+        if self.words is None:
+            self.words = " ".join(self.all_words[:self.num_start_words])
+            self.all_words = self.all_words[self.num_start_words:]
+        else:
+            # add words_per_epoch new words
+            count = 0
+            while count < self.words_per_epoch and len(self.all_words) > 0:
+                new_word = self.all_words[0]
+                self.words = " ".join(self.words.split(" ") + [new_word])
+                self.all_words = self.all_words[1:]
+                count += 1
+                # TODO: possibly do not increase count for stop-words and break if a "." is encountered.
+            # remove words until it fits in context length
+            while len(self.words) > perceptor.context_length:
+                # remove first word
+                self.words = " ".join(self.words.split(" ")[1:])
+        # get new encoding
+        print("Now thinking of: ", '"', self.words, '"')
+        sequence_number = self.get_img_sequence_number(epoch, iteration)
+        # save new words to disc
+        with open("story_transitions.txt", "a") as f:
+            f.write(f"{epoch}, {sequence_number}, {self.words}\n")
+        
+        encoding = self.create_text_encoding(self.words)
+        return encoding
 
     def image_output_path(self, sequence_number=None):
         """
@@ -353,37 +370,45 @@ class Imagine(nn.Module):
         if self.save_date_time:
             current_time = datetime.now().strftime("%y%m%d-%H%M%S_%f")
             output_path = f"{current_time}_{output_path}"
-        return Path(f"{output_path}.png")
+        return Path(f"{output_path}.jpg")
 
     def train_step(self, epoch, iteration):
         total_loss = 0
 
         for _ in range(self.gradient_accumulate_every):
             with autocast():
-                loss = self.model(self.clip_encoding)
+                out, loss = self.model(self.clip_encoding)
             loss = loss / self.gradient_accumulate_every
             total_loss += loss
-            self.scaler.scale(loss).backward()
-
+            self.scaler.scale(loss).backward()    
+        out = out.cpu().float().clamp(0., 1.)
         self.scaler.step(self.optimizer)
         self.scaler.update()
         self.optimizer.zero_grad()
 
         if (iteration % self.save_every == 0) and self.save_progress:
-            self.save_image(epoch, iteration)
+            self.save_image(epoch, iteration, img=out)
 
-        return total_loss
-
-    @torch.no_grad()
-    def save_image(self, epoch, iteration):
+        return out, total_loss
+    
+    def get_img_sequence_number(self, epoch, iteration):
         current_total_iterations = epoch * self.iterations + iteration
         sequence_number = current_total_iterations // self.save_every
+        return sequence_number
 
-        img = normalize_image(self.model(self.clip_encoding, return_loss=False).cpu())
-        img.clamp_(0., 1.)
+    @torch.no_grad()
+    def save_image(self, epoch, iteration, img=None):
+        sequence_number = self.get_img_sequence_number(epoch, iteration)
+
+        if img is None:
+            img = self.model(self.clip_encoding, return_loss=False).cpu().float().clamp(0., 1.)
         self.filename = self.image_output_path(sequence_number=sequence_number)
-        save_image(img, self.filename)
-        save_image(img, f"{self.textpath}.png")
+        
+        pil_img = T.ToPILImage()(img.squeeze())
+        pil_img.save(self.filename, quality=95, subsampling=0)
+        pil_img.save(f"{self.textpath}.jpg", quality=95, subsampling=0)
+        #save_image(img, self.filename)
+        #save_image(img, f"{self.textpath}.png")
 
         tqdm.write(f'image updated at "./{str(self.filename)}"')
 
@@ -409,7 +434,8 @@ class Imagine(nn.Module):
 
         tqdm.write(f'Imagining "{self.textpath}" from the depths of my weights...')
 
-        self.model(self.clip_encoding, dry_run = True) # do one warmup step due to potential issue with CLIP and CUDA
+        with torch.no_grad():
+            self.model(self.clip_encoding, dry_run=True) # do one warmup step due to potential issue with CLIP and CUDA
 
         if self.open_folder:
             open_folder('./')
@@ -418,11 +444,14 @@ class Imagine(nn.Module):
         for epoch in trange(self.epochs, desc='epochs'):
             pbar = trange(self.iterations, desc='iteration')
             for i in pbar:
-                loss = self.train_step(epoch, i)
+                _, loss = self.train_step(epoch, i)
                 pbar.set_description(f'loss: {loss.item():.2f}')
 
                 if terminate:
                     print('interrupted by keyboard, gracefully exiting')
                     return
+            # Update clip_encoding per epoch if we are creating a story
+            if self.create_story:
+                self.clip_encoding = self.update_story_encoding(epoch, i)
 
-        self.save_image(self.epochs, self.iterations) # one final save at end
+        self.save_image(epoch, i) # one final save at end
